/***************************************************************************
                          CondVar.h  -  description
                             -------------------
   Provides a portable conditional variable API.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __CONDVAR_H__
#define __CONDVAR_H__

#include "Mutex.h"

#if defined(POSIX)
/************************Start POSIX Version************************/
# include <stdlib.h>
# include <pthread.h>

/* Signals a change in the conditional variable.
   Returns non-zero value for success. */
#define signal_CONDVAR(cv) \
   (!pthread_cond_signal((pthread_cond_t *)cv))
/* Broadcasts a change in the conditional variable.
   Returns non-zero value for success. */
#define broadcast_CONDVAR(cv) \
   (!pthread_cond_broadcast((pthread_cond_t *)cv))
/* Waits for the conditional variable to change.
   Thread must hold a lock on the MUTEX m.
   Returns non-zero value for success. */
#define wait_CONDVAR(cv, m) \
   (!pthread_cond_wait((pthread_cond_t *)cv, (pthread_mutex_t *)m))
/* Destroys the conditional variable. */
#define del_CONDVAR(cv) \
   if (cv) { \
      pthread_cond_destroy((pthread_cond_t *)cv); \
      free(cv); \
   }

/*************************End POSIX Version*************************/
#elif defined(WIN32)
/************************Start WIN32 Version************************/
# ifndef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif
# include <windows.h>

/* Private Structure */
typedef struct {
   HANDLE event[2];
} WIN32_cond_t;

/* Signals a change in the conditional variable.
   Returns non-zero value for success. */
#define signal_CONDVAR(cv) \
   (PulseEvent(((WIN32_cond_t *)cv)->event[0]))
/* Broadcasts a change in the conditional variable.
   Returns non-zero value for success. */
#define broadcast_CONDVAR(cv) \
   (PulseEvent(((WIN32_cond_t *)cv)->event[1]))
/* Destroys the conditional variable. */
#define del_CONDVAR(cv) \
   if (cv) { \
      CloseHandle(((WIN32_cond_t *)cv)->event[0]); \
      CloseHandle(((WIN32_cond_t *)cv)->event[1]); \
      free(cv); \
   }

/*************************End WIN32 Version*************************/
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef void * CONDVAR;

/* Public Functions */
/* Creates a new conditional variable. */
CONDVAR new_CONDVAR();

/* The rest of these functions are implemented using the macros above. */
#ifdef IGNORE_THESE
/* Signals a change in the conditional variable.
   Returns non-zero value for success. */
int     signal_CONDVAR(CONDVAR cv);
/* Broadcasts a change in the conditional variable.
   Returns non-zero value for success. */
int     broadcast_CONDVAR(CONDVAR cv);
#endif
#ifdef WIN32
/* Waits for the conditional variable to change.
   Thread must hold a lock on the MUTEX m.
   Returns non-zero value for success. */
int     wait_CONDVAR(CONDVAR cv, MUTEX m);
#endif
#ifdef IGNORE_THESE
/* Destroys the conditional variable. */
void    del_CONDVAR(CONDVAR cv);
#endif

#ifdef __cplusplus
};
#endif

#endif
