/***************************************************************************
                        alsa_sound.h  -  description
                             -------------------
   Sound module implemented using ALSA.
                             -------------------
    begin                : Tue Mar 14 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#include <stdio.h>*/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <alsa/asoundlib.h>

#ifdef _SPEED_TEST_
# include <stdio.h>
static int frame_cnt = 0;
#endif

#define  REQ_PERIODS         32
#define  REQ_PERSIZE       1024
#define  REQ_BUFSIZE      (REQ_PERIODS * REQ_PERSIZE)
#ifdef _NO_51_SURROUND_
# define MAX_CHANNELS      2
#else
# define MAX_CHANNELS      6
#endif
#ifndef ALSA_DEFAULT_DEVICE
# define ALSA_DEFAULT_DEVICE   "plughw:0,0"
#endif

const char               *ALSA_DEVICE = ALSA_DEFAULT_DEVICE;
static snd_pcm_t         *alsa        = NULL;
static u32                channels    = 0,
                          rate        = 0,
                          buf_size    = 0;
static snd_pcm_uframes_t  buf_used    = 0;
static s16               *buf         = NULL;

/* Prepares the sound device for the incoming audio stream.
      INPUT: num_ch      - Number of audio channels.
             sample_rate - Sample rate (Hz)
      RETURN: Success - Number of audio channels.
              Failure - 0 */
u32 SOUND_init(u32 num_ch, u32 sample_rate)
{
#ifndef _SPEED_TEST_
   snd_pcm_hw_params_t *hw;
   unsigned int         set_rate    = sample_rate;
   snd_pcm_uframes_t    set_bufsize = REQ_BUFSIZE >> 1,
                        set_persize = REQ_PERSIZE >> 1;

   if (num_ch > 2)
      num_ch = MAX_CHANNELS;

   if (alsa) {
      if (!num_ch)
         goto SOUND_init_ERROR;
      else if ( (num_ch == channels)&&(sample_rate == rate) )
         return num_ch;
      snd_pcm_close(alsa);
      alsa = NULL;
   }
   else if (!num_ch)
      return 0;

   snd_pcm_hw_params_alloca(&hw);
   if (!hw)
      goto SOUND_init_ERROR;

   if (snd_pcm_open(&alsa, ALSA_DEVICE, SND_PCM_STREAM_PLAYBACK, 0) < 0)
      return 0;
   else if (snd_pcm_hw_params_any(alsa, hw) < 0)
      goto SOUND_init_ERROR;

   snd_pcm_hw_params_set_rate_resample(alsa, hw, 1);
   if ( (snd_pcm_hw_params_set_access(alsa, hw, SND_PCM_ACCESS_RW_INTERLEAVED) < 0)||
        (snd_pcm_hw_params_set_format(alsa, hw, SND_PCM_FORMAT_S16) < 0) )
      goto SOUND_init_ERROR;

   if (snd_pcm_hw_params_set_channels(alsa, hw, num_ch) < 0) {
      if (num_ch <= 2)
         goto SOUND_init_ERROR;     /* cannot set channels */
      else {
         /* cannot use 5.1 surround, so we're forced to live with stereo */
         num_ch = 2;
         if (snd_pcm_hw_params_set_channels(alsa, hw, num_ch) < 0)
            goto SOUND_init_ERROR;  /* cannot set channels */
      }
   }

   set_bufsize /= num_ch;
   set_persize /= num_ch;
   if ( (snd_pcm_hw_params_set_rate_near(alsa, hw, &set_rate, 0) < 0)||
        (set_rate != sample_rate)||
        (snd_pcm_hw_params_set_buffer_size_near(alsa, hw, &set_bufsize) < 0)||
        (snd_pcm_hw_params_set_period_size_near(alsa, hw, &set_persize, 0) < 0)||
        (snd_pcm_hw_params(alsa, hw) < 0) )
      goto SOUND_init_ERROR;
#endif

/*   printf("SOUND_init(%u, %u)\n", num_ch, sample_rate);*/
   channels = num_ch;
   rate     = sample_rate;
   return num_ch;

#ifndef _SPEED_TEST_
SOUND_init_ERROR:
   snd_pcm_close(alsa);
   alsa     = NULL;
   channels = rate = 0;
   return 0;
#endif
}

/* Returns the buffer that will hold the audio samples. Must
     call SOUND_init() function before calling this one.
      INPUT: samples - Number of samples the buffer will hold.
      RETURN:  Success - Pointer to the buffer.
               Failure - NULL */
s16 *SOUND_buf(u32 samples)
{
/*   printf("SOUND_buf(%u)", samples);*/
#ifndef _SPEED_TEST_
   if ( (alsa)&&(channels) ) {
#else
   if (channels) {
#endif
      buf_used = samples;
      samples *= sizeof(s16) * channels;
      if (samples > buf_size) {
         /* create a bigger buffer */
         if (buf)
            free(buf);
         buf      = malloc((size_t)samples);
         buf_size = (buf) ? samples : 0;
      }

      if (buf)
         memset(buf, 0, (size_t)samples);
      else
         buf_used = 0;
/*      printf("  buf_size=%u  buf_used=%u\n", buf_size, buf_used);*/
      return buf;
   }
   return NULL;
}

/* Called when the sound is no longer needed. */
void SOUND_done()
{
/*   printf("SOUND_done()\n");*/
#ifdef _SPEED_TEST_
   if (frame_cnt > 0) {
      printf("AUDIO: FRAMES=%d  TARGET_FRAME_RATE=%.3f\n", 
             frame_cnt, ((double)rate / (double)buf_used));
      frame_cnt = 0;
      channels  = rate = 0;
   }
#endif
   if (alsa) {
      snd_pcm_drain(alsa);
      snd_pcm_close(alsa);
      alsa     = NULL;
      channels = rate = 0;
   }
   if (buf) {
      free(buf);
      buf      = NULL;
      buf_size = 0;
   }
}

/* Called when the samples in the buffer should be played. */
void SOUND_play()
{
#ifndef _SPEED_TEST_
   if (snd_pcm_writei(alsa, buf, buf_used) < 0)
      snd_pcm_prepare(alsa);
#else
   frame_cnt++;
#endif
}

/* Returns the number of samples left in the buffer. */
u32 SOUND_samples_left()
{
   snd_pcm_sframes_t samples;

   if ( (snd_pcm_delay(alsa, &samples) < 0)||(samples <= 0) )
      return 0;   /* don't know how many bytes are left */
   return (u32)samples;
}

void SOUND_stop(wait)
{
   snd_pcm_drop(alsa);
}

#undef REQ_PERIODS
#undef REQ_PERSIZE
#undef REQ_BUFSIZE
#undef MAX_CHANNELS

